#!/usr/bin/env python3
# coding: utf-8

import os
import sys
import argparse
import polib
import shutil
import re
from msg2po.core import (
    VALID_EXTENSIONS,
    dir_or_exit,
    get_enc,
    cd,
    basename,
    metadata,
    file2po,
    get_ext,
    po_make_unique,
    sort_po,
    parent_dir,
    is_indexed,
    CONFIG,
)
import natsort


# prepare po dir
def prepare_po_dir(d):
    if not os.path.isdir(d):
        shutil.rmtree(d, ignore_errors=True)
        os.makedirs(d)
    tmp_dir = os.path.join(d, "tmp")
    if not os.path.isdir(tmp_dir):
        shutil.rmtree(tmp_dir, ignore_errors=True)
        os.makedirs(tmp_dir)


def clean_po_dir(d):
    tmp_dir = os.path.join(d, "tmp")
    shutil.rmtree(tmp_dir, ignore_errors=True)


def poify(poify_dir: str, encoding: str = CONFIG.encoding):
    """
    poify_dir is path to source language directory
    """
    language_dir = os.path.basename(poify_dir)
    po_dir = CONFIG.po_dirname
    prepare_po_dir(po_dir)
    tra_dir = tra_relpath(poify_dir)
    # process with po_tool
    with cd(language_dir):
        # Final PO
        lang = basename(language_dir)
        dst_file = os.path.join(po_dir, lang + ".pot")
        po = polib.POFile()

        # skip female cuts, they are built from male ones
        extract_format = CONFIG.extract_format

        skip_files = CONFIG.skip_files

        for dir_name, subdir_list, file_list in natsort.natsorted(
            os.walk(".", topdown=False, followlinks=True), alg=natsort.PATH
        ):
            subdir_list = natsort.natsorted(subdir_list, alg=natsort.PATH)
            file_list = natsort.natsorted(file_list, alg=natsort.PATH)
            for file_name in file_list:
                full_name = os.path.join(dir_name, file_name)
                full_name = re.sub("^\./", "", full_name)  # remove trailing './'

                # skip female cuts
                pretty_dir_name = re.sub("^\./", "", dir_name)
                if extract_format == "sfall" and pretty_dir_name == "cuts_female":
                    print("{} is in cuts_female. Skipping!".format(full_name))
                    continue

                if full_name in skip_files:
                    print("{} is in skip_files. Skipping!".format(full_name))
                    continue

                ext = get_ext(file_name)
                if ext not in VALID_EXTENSIONS:
                    continue

                # checked txt is indexed and if it is, process it
                if ext == "txt":
                    if is_indexed(full_name):
                        print("{} is indexed TXT".format(full_name))
                    else:
                        print("{} is TXT, but not indexed. Skipping!".format(full_name))
                        continue

                bname = basename(full_name)
                # non-default encoding?
                if encoding == CONFIG.encoding:
                    enc = get_enc(file_path=bname)
                else:
                    enc = encoding
                print("processing {} with encoding {}".format(full_name, enc))
                po2 = file2po(full_name, encoding=enc)
                for e2 in po2:
                    po.append(e2)
    po = po_make_unique(po)
    po = sort_po(po)
    clean_po_dir(po_dir)

    old_po = polib.pofile(dst_file)
    po.metadata = old_po.metadata
    if po == old_po:
        print("No change in source directory {}".format(poify_dir))
        sys.exit(0)
    else:
        po.metadata = metadata(pot=True)

    po.save(dst_file, newline=CONFIG.newline_po)

    print("Processed directory {}, the result is in {}/{}/{}.pot".format(poify_dir, tra_dir, po_dir, lang))


def tra_relpath(poify_dir: str) -> str:
    return os.path.relpath(parent_dir(poify_dir))


def main():
    # parse args
    parser = argparse.ArgumentParser(
        description="Poify files in selected directory", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "DIR",
        nargs="?",
        default="{}".format(CONFIG.poify_dir),
        help="source language directory",
    )
    parser.add_argument("-e", dest="enc", help="source encoding", default="{}".format(CONFIG.encoding))
    args = parser.parse_args()

    # init vars
    poify_dir = args.DIR
    dir_or_exit(poify_dir)

    # so that resulting po has relative occurences
    with cd(parent_dir(os.path.abspath(poify_dir))):
        poify(poify_dir)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8

import argparse
from msg2po.core import po2file, VALID_EXTENSIONS
from polib import pofile

formats = "/".join(VALID_EXTENSIONS)
parser = argparse.ArgumentParser(
    description="Extract {} from gettext PO".format(formats), formatter_class=argparse.ArgumentDefaultsHelpFormatter
)
parser.add_argument("INPUT_FILE", help="input PO file")
parser.add_argument("OUTPUT_FILE", help="output translation file")
parser.add_argument("-e", dest="encoding", help="output encoding", default="cp1252")
parser.add_argument("--path", dest="path", help="file occurrence in PO (relative path) - defaults to output name")
args = parser.parse_args()

if args.path is None:
    path = args.OUTPUT_FILE
else:
    path = args.path


def main():
    po = pofile(args.INPUT_FILE)
    po2file(po, args.OUTPUT_FILE, args.encoding, path)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8

import argparse
from msg2po.core import file2msgstr, CONFIG, VALID_EXTENSIONS
from polib import pofile

formats = "/".join(VALID_EXTENSIONS)
parser = argparse.ArgumentParser(
    description="Load {} into gettext PO msgstr".format(formats), formatter_class=argparse.ArgumentDefaultsHelpFormatter
)
parser.add_argument("INPUT_FILE", help="input file")
parser.add_argument("OUTPUT_FILE", help="output PO file")
parser.add_argument(
    "--path", dest="path", help="load as filename (PO occurrence). If not specified, defaults to input filename"
)
parser.add_argument("-e", dest="encoding", default="cp1252", help="source encoding")
parser.add_argument(
    "overwrite", dest="overwrite", default=False, action="store_true", help="overwrite existing msgstr's"
)
parser.add_argument(
    "--same",
    dest="same",
    default=False,
    action="store_true",
    help="load translations that are identical to original strings",
)
args = parser.parse_args()

input_file = args.INPUT_FILE
output_file = args.OUTPUT_FILE
if args.path is None:
    path = args.INPUT_FILE
else:
    path = args.path


def main():
    po = pofile(output_file)
    po = file2msgstr(input_file, po, path, encoding=args.encoding, overwrite=args.overwrite, same=args.same)
    po.save(output_file, newline=CONFIG.newline_po)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import os
import argparse
import sys

# po: new translations added through weblate use case sensitive code: pt_BR.po. Keeping them.
EXCLUDE_FILES = ["README.md"]
EXCLUDE_DIRS = [".git", ".svn", ".hg", ".github"]

# parse args
parser = argparse.ArgumentParser(
    description="Lowercase files in selected directory", formatter_class=argparse.ArgumentDefaultsHelpFormatter
)
parser.add_argument("dir")
args = parser.parse_args()


# https://stackoverflow.com/questions/3075443/python-recursively-remove-capitalisation-from-directory-structure
def lowercase_rename(dir):
    def rename_all(root, items):
        for name in items:
            new_name = name.lower()
            if name == new_name:
                continue
            if (name in EXCLUDE_FILES) or name.endswith(".po"):
                print("{}: reserved name, skip".format(name))
                continue
            path = os.path.join(root, name)
            new_path = os.path.join(root, new_name)

            full_path = os.path.normpath(path)
            split_path = full_path.split(os.sep)
            if len((set(EXCLUDE_DIRS) & set(split_path))) > 0:
                print("{}: reserved dir name in path, skip".format(path))
                continue

            print("renaming {} to {}".format(path, new_path))
            os.rename(path, new_path)

    # starts from the bottom so paths further up remain valid after renaming
    for root, dirs, files in os.walk(dir, topdown=False):
        rename_all(root, dirs)
        rename_all(root, files)


def main():
    if not os.path.isdir(args.dir):
        print("Error: {} is not a directory".format(args.dir))
        sys.exit(1)
    lowercase_rename(args.dir)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8

import argparse
from msg2po.core import CONFIG, sort_po, unfuzzy_exact_matches
import csv
from collections import OrderedDict
import polib

parser = argparse.ArgumentParser(
    description="Load female csv into Gettext PO", formatter_class=argparse.ArgumentDefaultsHelpFormatter
)
parser.add_argument("INPUT_CSV", help="input CSV file")
parser.add_argument("OUTPUT_PO", help="output PO file")
args = parser.parse_args()


def load_csv(path):
    female_strings = OrderedDict()
    with open(path, "r") as csvfile:
        reader = csv.reader(csvfile)
        for row in reader:
            female_strings[row[0]] = row[1]
    sorted_strings = OrderedDict(sorted(female_strings.items()))
    return sorted_strings


def main():
    female_strings = load_csv(args.INPUT_CSV)

    po = polib.pofile(args.OUTPUT_PO)
    for fs in female_strings:
        entry = polib.POEntry(msgid=fs, msgstr=female_strings[fs], msgctxt="female")
        po.append(entry)
    po = sort_po(po)
    po = unfuzzy_exact_matches(po)
    po.save(args.OUTPUT_PO, newline=CONFIG.newline_po)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8

import os
import argparse
import re
import sys
from msg2po.core import (
    VALID_EXTENSIONS,
    LanguageMap,
    cd,
    female_entries,
    find_files,
    get_ext,
    get_enc,
    file2msgstr,
    po_make_unique,
    CONFIG,
    basename,
)
from polib import pofile, POFile

# parse args
parser = argparse.ArgumentParser(
    description="Load strings from files in selected dir into PO msgstr's",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument(
    "--auto",
    dest="auto",
    default=False,
    action="store_true",
    help="automatically find POs and language dirs, process all valid extensions",
    required=False,
)
parser.add_argument("-s", dest="src_dir", default=".", help="directory to load", required=False)
parser.add_argument("-o", dest="output_file", default=None, help="existing PO file", required=False)
parser.add_argument("--ext", dest="file_ext", default=None, help="load files with this extension", required=False)
parser.add_argument(
    "--same",
    dest="same",
    default=False,
    action="store_true",
    help="load translations that are identical to original strings",
)
parser.add_argument(
    "--overwrite", dest="overwrite", default=False, action="store_true", help="overwrite existing translations"
)
args = parser.parse_args()


def dir2msgstr(src_dir: str, po: POFile, po_path: str = "", overwrite: bool = True, extension: str = ""):
    """
    src_dir is relative
    overwrite means ovewrite existing entries if any
    """
    print("overwrite is " + str(overwrite))

    with cd(src_dir):
        female_map = female_entries(po)

        for dir_name, subdir_list, file_list in os.walk(".", topdown=False, followlinks=True):
            for file_name in file_list:
                full_name = os.path.join(dir_name, file_name)
                full_name = re.sub("^\./", "", full_name)  # remove trailing './'
                fext = get_ext(file_name)
                if fext != extension:
                    continue
                if dir_name.endswith(CONFIG.female_dir_suffix):
                    print("{} is a file with female strings, skipping".format(full_name))
                    continue

                enc = get_enc(po_path, file_name)
                print("processing {} with encoding {}".format(full_name, enc))
                po = file2msgstr(
                    input_file=full_name,
                    po=po,
                    occurence_path=full_name,
                    encoding=enc,
                    overwrite=overwrite,
                    same=args.same,
                    female_map=female_map
                )
    po = po_make_unique(po)
    return po


def main():
    if not args.auto and ((args.output_file is None) or (args.file_ext is None)):
        print("ERROR: you must either use auto mode or specify output PO and file extension")
        sys.exit(1)

    if not args.auto:
        output_file = args.output_file
        po = pofile(output_file)
        po = dir2msgstr(
            src_dir=args.src_dir, po=po, po_path=output_file, overwrite=args.overwrite, extension=args.file_ext
        )
        po.save(output_file, newline=CONFIG.newline_po)
        print("Processed directory {}, the result is in {}".format(args.src_dir, output_file))

    if args.auto:
        language_map = LanguageMap()
        with cd(CONFIG.tra_dir):
            po_paths = find_files(CONFIG.po_dirname, "po")
            for pf in po_paths:
                print("Loading into {}".format(pf))
                lang_dir = language_map.po2slug[basename(pf)]
                po = pofile(pf)
                for ve in VALID_EXTENSIONS:
                    po = dir2msgstr(src_dir=lang_dir, po=po, po_path=pf, overwrite=args.overwrite, extension=ve)
                    po.save(pf, newline=CONFIG.newline_po)
                    print("Processed {} files in directory {}, the result is in {}".format(ve, lang_dir, pf))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import polib

parser = argparse.ArgumentParser(
    description="Unfuzzy Cassidy lines which were fuzzied by adding sounds",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument("INPUT_FILE", help="input PO file")
parser.add_argument("-w", default=False, dest="WRITE", action="store_true", help="save PO file?")
args = parser.parse_args()

input_file = args.INPUT_FILE
write = args.WRITE
po = polib.pofile(input_file)

i = 0
for e in po.fuzzy_entries():
    if e.previous_msgid:  # some fuzzies are assigned automatically on merge and don't have previous msgids
        ctxt = e.msgctxt
        if ctxt and ctxt.startswith("Cas") and (e.msgid != e.msgstr):
            if write:
                e.flags.remove("fuzzy")
                e.previous_msgid = None
                e.previous_msgid_plural = None
                e.previous_msgctxt = None
            else:  # preview
                i = i + 1
                print(e.msgctxt)
                print(e.msgid)
                print(e.msgstr)

if write:
    po.save(input_file)
else:
    print(i)



#!/usr/bin/env python3
# coding: utf-8

# this script is needed for shell wrappers

import sys
from msg2po.config import CONFIG
import ruamel.yaml

stanza = sys.argv[1]
if stanza != "dorny-paths":
    key = sys.argv[2]


def main():
    if stanza == "dorny-paths":
        dorny_paths()
        sys.exit(0)
    try:
        value = CONFIG._config[stanza][key]
        print(value)
    except:
        print("config {}:{} not found".format(stanza, key))
        sys.exit(1)


# for github action https://github.com/dorny/paths-filter
def dorny_paths():
    yaml = ruamel.yaml.YAML()
    tra_dir = CONFIG.tra_dir
    src_lang = CONFIG.src_lang
    unpoify = ["{}/po/*.po".format(tra_dir)]
    if CONFIG.extract_format == "sfall":  # fallout
        poify = [
            "{}/{}/*/*.msg".format(tra_dir, src_lang),
            "{}/{}/*/*.sve".format(tra_dir, src_lang),
            "{}/{}/*/*.txt".format(tra_dir, src_lang),
        ]
        dir2msgstr = [
            "{}/*/*/*.msg".format(tra_dir),
            "{}/*/*/*.sve".format(tra_dir),
            "{}/*/*/*.txt".format(tra_dir),
            "!{}/{}/*/*.msg".format(tra_dir, src_lang),
            "!{}/{}/*/*.sve".format(tra_dir, src_lang),
            "!{}/{}/*/*.txt".format(tra_dir, src_lang),
        ]
    else:
        poify = ["{}/{}/**/*.tra".format(tra_dir, src_lang)]
        dir2msgstr = [
            "{}/**/*.tra".format(tra_dir),
            "!{}/{}/**/*.tra".format(tra_dir, src_lang),
        ]
    paths = {"poify": poify, "unpoify": unpoify, "dir2msgstr": dir2msgstr}
    yaml.dump(paths, sys.stdout)


if __name__ == "__main__":
    main()



# functions that are needed in config and core, to avoid circular import
import os


def find_files(dir: str, ext: str):
    """
    Find files with extension ext in directory dir
    """
    files = []
    for root, subdir_list, file_list in os.walk(dir):
        for f in file_list:
            if get_ext(f) == ext:
                files.append((os.path.join(root, f)))
    return files


def get_ext(path):
    try:
        ext = path.rsplit(".", 1)[1].lower()
    except:
        ext = None
    return ext



#!/usr/bin/env python3
# coding: utf-8
"""
Update POs from POT. Behaves like
msgmerge -v --previous --no-wrap -U -N --backup=off lang1.po lang0.pot
Except that it allows string with "female" context to remain in PO,
if there's a corresponding male entry without such context.
"""

from functools import partial
import os
import argparse
import subprocess
import sys
from multiprocessing import Pool
from msg2po.core import sort_po, update_female_entries, CONFIG, find_files, unfuzzy_exact_matches
from polib import pofile

# parse args
parser = argparse.ArgumentParser(
    description="Update POs from POT, keeping female entries. Requires Gettext msgmerge in PATH",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)

parser.add_argument("PO", help="PO file", nargs="?", default=None)
parser.add_argument("POT", help="POT file", nargs="?", default=None)
args = parser.parse_args()


def merge(po_path: str, pot_path: str):
    print(po_path)
    exit_code = 0
    cmd = ["msgmerge", "--previous", "--no-wrap", "-U", "-q", "--backup=off", po_path, pot_path]
    res = subprocess.run(
        cmd,
        capture_output=True,
        check=True,
        text=True,
    )
    print(res.stdout)
    print(res.stderr)
    if res.returncode != 0:
        exit_code = res.returncode
        print("ERROR: msgmerge failed for {}".format(po_path))
    po2 = pofile(po_path)
    po2 = update_female_entries(po2)
    po2 = sort_po(po2)
    po2 = unfuzzy_exact_matches(po2)
    po2.save(fpath=po_path, newline=CONFIG.newline_po)
    return exit_code


def main():
    # single file
    if (args.PO is not None) and (args.POT is not None):
        res = merge(args.PO, args.POT)
        if res != 0:
            print("ERROR: msgmerge failed for {}".format(args.PO))
            sys.exit(1)
        sys.exit(0)

    # multifile, read .bgforge.yml
    po_dir = CONFIG.po_dir
    po_files = find_files(po_dir, "po")
    pot_file = os.path.join(po_dir, CONFIG.src_lang + ".pot")

    print("Merging PO files in {} with {}".format(po_dir, pot_file))
    pool = Pool()
    try:
        r = pool.map_async(partial(merge, pot_path=pot_file), po_files)
        pool.close()
        codes = r.get()
    except KeyboardInterrupt:
        pool.terminate()
    finally:
        pool.join()

    for c in codes:
        if c != 0:
            print("ERROR: one of msgmerge invocations failed.")
            sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8

import argparse
from polib import pofile
from msg2po.core import sort_po, update_female_entries, CONFIG, unfuzzy_exact_matches

parser = argparse.ArgumentParser(
    description="Resave PO file using polib API, to correct formatting",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument("INPUT_FILE", help="PO file to resave")
args = parser.parse_args()


def main():
    po = pofile(args.INPUT_FILE)
    po = update_female_entries(po)
    po2 = sort_po(po)
    po2 = unfuzzy_exact_matches(po2)
    po2.save(args.INPUT_FILE, newline=CONFIG.newline_po)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8

import os
import argparse
import sys
from multiprocessing import Pool
from polib import pofile
from functools import partial
from msg2po.core import (
    CONFIG,
    LanguageMap,
    dir_or_exit,
    cd,
    female_entries,
    translation_entries,
    get_enc,
    po2file,
    get_ext,
)

# parse args
parser = argparse.ArgumentParser(
    description="Unpoify files into parent directory of selected PO dir",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument("DIR", nargs="?", default=CONFIG.po_dir, help="directory with PO files")
args = parser.parse_args()

# init vars
po_dir = args.DIR
dir_or_exit(po_dir)


def extract_po(pf: str, language_map: LanguageMap):
    """
    pf is po file basename
    """
    po_path = os.path.join(CONFIG.po_dirname, pf)
    print("processing {}".format(po_path))
    po = pofile(po_path)  # open once
    trans_map = translation_entries(po)
    female_map = female_entries(po)

    dst_dir = language_map.po2slug[pf]

    for ef in sorted(trans_map):
        enc = get_enc(po_path, ef)
        ef_extract_path = os.path.join(dst_dir, ef)
        print("Extracting {} from {} into {} with encoding {}".format(ef, po_path, ef_extract_path, enc))
        po2file(po, ef_extract_path, enc, ef, dst_dir=dst_dir, trans_map=trans_map, female_map=female_map)

    print("Extracted {} into {} with encoding {}".format(po_path, dst_dir, enc))


def main():
    po_dir = CONFIG.po_dir
    language_map = LanguageMap()

    # find PO files
    po_files = []
    for dir_name, subdir_list, file_list in os.walk(po_dir):
        for f in file_list:
            if get_ext(f) == "po":
                po_files.append(f)
    if po_files == []:
        print("no PO files found in directory {}".format(po_dir))
        sys.exit(1)

    with cd(CONFIG.tra_dir):
        # extract PO files
        pool = Pool()
        try:
            r = pool.map_async(partial(extract_po, language_map=language_map), po_files)
            pool.close()
            codes = r.get()  # noqa: F841 - need to get results to see if there's an exception
        except KeyboardInterrupt:
            pool.terminate()
        finally:
            pool.join()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8

import argparse
from msg2po.core import file2po, CONFIG, VALID_EXTENSIONS

formats = "/".join(VALID_EXTENSIONS)
parser = argparse.ArgumentParser(
    description="Load {} into gettext PO msgid's".format(formats),
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument("INPUT_SOURCE", help="input source file")
parser.add_argument("OUTPUT_PO", help="output PO file")
parser.add_argument("-e", dest="enc", default="{}".format(CONFIG.encoding), help="source encoding")
args = parser.parse_args()


def main():
    po = file2po(args.INPUT_SOURCE, args.OUTPUT_PO, args.enc)
    po.save(args.OUTPUT_PO, newline=CONFIG.newline_po)


if __name__ == "__main__":
    main()



import ruamel.yaml
import sys
import os

VERSION = "1.1.12"


class Config:
    def __init__(self):
        yml = ".bgforge.yml"
        translation_defaults = {
            "encoding": "cp1252",
            "tra_dir": ".",
            "src_lang": "english",
            "simple_languages": True,  # extract into language name, not language code. pt_BR.po -> portuguese/1.msg
            "skip_files": [],
            "extract_format": "",  # could be 'sfall'
            "no_female": False,  # explicitly disable female extraction (TODO: reason unclear)
            "extract_fuzzy": False,
            "lowercase": True,
        }

        config = translation_defaults
        try:
            with open(yml) as yf:
                yaml = ruamel.yaml.YAML()
                config = yaml.load(yf)
            translation_config = {**translation_defaults, **config["translation"]}
        except:
            print(yml + " not found, assuming defaults", file=sys.stderr)
            translation_config = translation_defaults
        config["translation"] = translation_config
        self._config = config  # so that shell wrapper can get values from other stanzas too

        self.encoding = translation_config["encoding"]
        self.tra_dir = translation_config["tra_dir"]
        self.src_lang = translation_config["src_lang"]
        self.simple_languages = translation_config["simple_languages"]
        self.skip_files = translation_config["skip_files"]
        self.extract_format = translation_config["extract_format"]
        self.no_female = translation_config["no_female"]
        self.extract_fuzzy = translation_config["extract_fuzzy"]
        self.lowercase = translation_config["lowercase"]

        self.po_dirname = "po"
        self.female_dir_suffix = "_female"
        self.po_dir = os.path.join(self.tra_dir, self.po_dirname)
        self.poify_dir = os.path.join(self.tra_dir, self.src_lang)
        self.version = VERSION
        self.newline_tra = "\r\n"
        self.newline_po = "\n"


CONFIG = Config()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
from polib import pofile
import ruamel.yaml
import os
from msg2po.core import CONFIG

parser = argparse.ArgumentParser(
    description="Unmark PO entries as fuzzy,"
    "if replacing string 1 with string2 in previous msgid results in current msgid",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument("INPUT_FILE", help="input PO file")
parser.add_argument("-w", default=False, dest="WRITE", action="store_true", help="save PO file?")
args = parser.parse_args()

input_file = args.INPUT_FILE
write = args.WRITE
po = pofile(input_file)


def load_replacements():
    if os.path.isfile("unfuzzy.yml"):
        yml = "unfuzzy.yml"
    else:
        yml = os.path.join(os.path.dirname(os.path.realpath(__file__)), "unfuzzy.yml")
    with open(yml) as yf:
        yaml = ruamel.yaml.YAML()
        replace_list = yaml.load(yf)
    return replace_list


def make_replaces(line, replace_list):
    for r in replace_list:
        line = line.replace(r[0], r[1])
    return line


def msgids_equal(id1, id2, replace_list):
    id1 = make_replaces(id1, replace_list)
    id2 = make_replaces(id2, replace_list)
    if id1 == id2:
        return True
    else:
        return False


def main():
    i = 0
    replace_list = load_replacements()
    for e in po.fuzzy_entries():
        if e.previous_msgid:  # some fuzzies are assigned automatically on merge and don't have previous msgids
            e1 = e.msgid
            e2 = e.previous_msgid
            if msgids_equal(e1, e2, replace_list):
                if write:
                    e.flags.remove("fuzzy")
                    e.previous_msgid = None
                    e.previous_msgid_plural = None
                    e.previous_msgctxt = None
                else:  # preview
                    i = i + 1
                    print("OLD SOURCE:  {}".format(e1))
                    print("NEW SOURCE:  {}".format(e2))
                    print("TRANSLATION: {}".format(e.msgstr))
                    print("")

    if write:
        po.save(input_file, newline=CONFIG.newline_po)
    else:
        print(i)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# diffs male vs female PO and saves the result into CSV suitable for BGforge Weblate
# PO files MUST be identical - except for msgstr's (translations)

import polib
import csv
import sys


def main():
    male_po = polib.pofile(sys.argv[1])
    female_po = polib.pofile(sys.argv[2])
    female_csv = sys.argv[3]

    rows = []

    for i in range(0, len(male_po) - 1):
        if male_po[i].msgstr != female_po[i].msgstr:
            rows.append([male_po[i].msgid, female_po[i].msgstr])

    rows = sorted(rows, key=lambda k: k[0])

    with open(female_csv, "w") as csvfile:
        writer = csv.writer(csvfile, quoting=csv.QUOTE_ALL)
        writer.writerows(rows)


if __name__ == "__main__":
    main()



import re
import sys
from collections import OrderedDict
import polib
import os
import shutil
from contextlib import contextmanager
from natsort import natsorted
from msg2po.config import CONFIG
from datetime import datetime
from msg2po.common import find_files, get_ext

# extensions recognized by file2po, etc
VALID_EXTENSIONS = ["msg", "txt", "sve", "tra"]

# supported file formats
# pattern is used to parse original files
# line_format to write to translated files
# index, value, context, female - order of these tokens in pattern
# dotall - whether file entries are multiline
FILE_FORMAT = {
    "msg": {
        "pattern": "{(\d+)}{([^}]*)}{([^}]*)}",
        "dotall": True,
        "index": 0,
        "value": 2,
        "context": 1,
        "line_format": {
            "default": "{{{index}}}{{}}{{{value}}}\n",
            "context": "{{{index}}}{{{context}}}{{{value}}}\n",
            "female": "separate",
        },
    },
    "sve": {
        "pattern": "(\d+):(.*)",
        "dotall": False,
        "index": 0,
        "value": 1,
        "line_format": {
            "default": "{index}:{value}\n",
            "female": "separate",
        },
    },
    "txt": {
        "pattern": "(\d+):(.*)",
        "dotall": False,
        "index": 0,
        "value": 1,
        "comment": "indexed_txt",
        "line_format": {
            "default": "{index}:{value}\n",
            "female": "separate",
        },
    },
    "tra": {
        "pattern": "@(\d+)\s*?=\s*?~([^~]*?)~(?:\s)?(?:\[([^]]*)\])?(?:~([^~]*)~)?",
        "dotall": True,
        "index": 0,
        "value": 1,
        "context": 2,
        "female": 3,
        "line_format": {
            "default": "@{index} = ~{value}~\n",
            "context": "@{index} = ~{value}~ [{context}]\n",
            "female": "@{index} = ~{value}~ ~{female}~\n",
        },
    },
}


# used for determining empty strings, which are invalid by PO spec
EMPTY_COMMENT = "LEAVE empty space in translation"

CONTEXT_FEMALE = "female"


# file and dir manipulation
#################################


def basename(path):
    if path.endswith(os.sep):
        path = path[:-1]
    return os.path.abspath(path).rsplit(os.sep, 1)[1]


def parent_dir(path):
    if path.endswith(os.sep):
        path = path[:-1]
    return os.path.abspath(path).rsplit(os.sep, 1)[0]


def strip_ext(filename):
    return filename.rsplit(".", 1)[0]


def get_dir(path: str):
    return path.rsplit(os.sep, 1)[0]


def create_dir(path):
    if not os.path.isdir(path):
        os.makedirs(path)


def dir_or_exit(d):
    if os.path.isdir(d):
        print("Found directory {}".format(d))
    else:
        print("Directory {} does not exist, cannot continue!".format(d))
        sys.exit(1)


@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def get_enc(lang_path: str = "", file_path: str = ""):
    """
    Infers encoding based on dir/PO name and file path
    lang_path can be PO path or translation path, only basename is used
    """
    ENCODINGS = {
        "schinese": "cp936",
        "tchinese": "cp950",
        "czech": "cp1250",
        "hungarian": "cp1250",
        "japanese": "cp932",
        "korean": "cp949",
        "polish": "cp1250",
        "polski": "cp1250",
        "russian": "cp1251",
        "ukrainian": "cp1251",
    }

    DOS_ENCODINGS = {
        #  'czech': 'cp852',
        #  'polish': 'cp852',
        #  'polski': 'cp852',
        "russian": "cp866",
        "ukrainian": "cp866",
        #  'french': 'cp850',
        #  'francais': 'cp850',
        #  'german': 'cp850',
        #  'deutsch': 'cp850',
        #  'italian': 'cp850',
        #  'italiano': 'cp850',
        #  'spanish': 'cp850',
        #  'espanol': 'cp850',
        #  'castilian': 'cp850',
        #  'castellano': 'cp850',
    }

    DOS_FILENAMES = [
        "setup.tra",
        "install.tra",
    ]

    UTF_FILENAMES = [
        "ee.tra",
    ]

    encoding = CONFIG.encoding
    lang = language_slug(lang_path)
    filename = basename(file_path)

    if lang in ENCODINGS:
        encoding = ENCODINGS[lang]

    if filename in DOS_FILENAMES:
        try:
            encoding = DOS_ENCODINGS[lang]
        except:
            pass

    if filename in UTF_FILENAMES:
        encoding = "utf-8"

    utf_name = re.compile(".*_ee.tra$")
    if utf_name.match(filename):
        encoding = "utf-8"

    return encoding


################################


def metadata(old_metadata=None, pot=False, po=False):
    if old_metadata is None:
        data = {
            "Project-Id-Version": "PACKAGE VERSION",
            "Report-Msgid-Bugs-To": "",
            "Last-Translator": "FULL NAME <EMAIL@ADDRESS>",
            "Language-Team": "LANGUAGE <LL@li.org>",
            "Language": "",
            "MIME-Version": "1.0",
            "Content-Type": "text/plain; charset=UTF-8",
            "Content-Transfer-Encoding": "8bit",
            "X-Generator": "BGforge/msg2po v.{}".format(CONFIG.version),
        }
        if pot:
            data["POT-Creation-Date"] = datetime.today().strftime("%Y-%m-%d-%H:%M") + "+0000"
        if po:
            data["PO-Revision-Date"] = datetime.today().strftime("%Y-%m-%d-%H:%M") + "+0000"
    else:
        return old_metadata
    return data


def file2po(filepath: str, po_path: str = "", encoding=CONFIG.encoding):
    """Returns PO file object"""

    trans = TRANSFile(filepath=filepath, is_source=True, encoding=encoding)  # load translations

    if po_path == "":
        po = polib.POFile()
        po.metadata = metadata()
    else:
        po = polib.pofile(po_path)

    trans_map = {}
    i = 0  # index in PO object
    for t in trans.entries:
        context = t.context

        # append to occurrences if id and context match
        if (t.value, context) in trans_map:
            e = po[trans_map[(t.value, context)]]
            e.occurrences.append((filepath, t.index))
            continue

        # no matching msgid + msgctxt, add new entry
        entry = polib.POEntry(
            msgid=t.value,
            msgstr="",
            occurrences=[
                (filepath, t.index),
            ],
            msgctxt=t.context,
            comment=t.comment,
        )
        po.append(entry)
        trans_map[(t.value, context)] = i
        i = i + 1

    return po


# check if extract file is present in po, exit with error if not
def check_path_in_po(po, path):
    present_files = set()
    for entry in po:
        for eo in entry.occurrences:
            present_files.add(eo[0])
    present_files_list = sorted(set(present_files))
    if path not in present_files_list:
        print("{} is not present in selected PO file".format(path))
        print("supply one of present files with --path argument:")
        for pf in present_files_list:
            print(pf)
        sys.exit(1)


def translation_entries(po: polib.POFile):
    """
    returns {filepath: [{"file_index": index_in_file, "po_index": index_in_po}] }
    does not include female entries, as they don't have occurences
    """
    entries = {}
    i = 0
    for entry in po:
        for eo in entry.occurrences:
            path = eo[0]
            linenum = eo[1]
            if path in entries:
                entries[path].append({"file_index": int(linenum), "po_index": i})
            else:
                entries[path] = [{"file_index": int(linenum), "po_index": i}]
        i = i + 1
    return entries


def female_entries(po: polib.POFile) -> "dict[str: polib.POEntry]":
    """
    Returns mapping of male msgids to corresponding female PO entries
    """
    entries = {}
    fe_list = [e for e in po if len(e.occurrences) == 0 and e.msgctxt == CONTEXT_FEMALE]
    for fe in fe_list:
        # first, check male entries without context
        male_entries = [e for e in po if e.msgid == fe.msgid and not e.msgctxt]
        if len(male_entries) > 0:
            me = male_entries[0]
        else:  # then, those with
            male_entries = [e for e in po if e.msgid == fe.msgid and e.msgctxt != CONTEXT_FEMALE]
        try:
            me = male_entries[0]
            entries[me.msgid] = fe
        except:
            print("WARNING: couldn't find a corresponding male counterpart for a female entry")
            print(fe)
    return entries


def po2file(
    po: polib.POFile,
    output_file: str,
    encoding: str,
    occurrence_path: str,
    dst_dir=None,
    trans_map=None,
    female_map=None,
):
    """
    Extract and write to disk a single file from POFile
    output_file is path relative to dst_dir
    dst_dir is actually dst language. Used only in unpoify
    """
    if trans_map is None:  # when extracting single file with po2tra/po2msg, etc
        # check if file is present in po, exit if not
        check_path_in_po(po, occurrence_path)
        trans_map = translation_entries(po)
    if female_map is None:
        female_map = female_entries(po)

    # create parent directory
    create_dir(get_dir(output_file))

    ext = get_ext(output_file)
    ff = FILE_FORMAT[ext]
    line_format = ff["line_format"]

    context = ""
    resulting_entries = []
    extract_fuzzy = CONFIG.extract_fuzzy

    for file_trans in trans_map[occurrence_path]:
        file_index = file_trans["file_index"]
        po_index = file_trans["po_index"]
        entry = po[po_index]

        if entry.msgstr == "":  # if not translated, keep msgid
            value = entry.msgid
        elif "fuzzy" in entry.flags and not extract_fuzzy:  # skip fuzzy?
            value = entry.msgid
        else:
            value = entry.msgstr  # either translated or fuzzy+extract_fuzzy

        # empty lines detected by comment
        if entry.comment == EMPTY_COMMENT:
            value = ""

        # context
        context = entry.msgctxt

        # female strings
        female = None
        if entry.msgid in female_map:
            fe_entry = female_map[entry.msgid]
            if fe_entry.msgstr == "":
                female = fe_entry.msgid
            elif "fuzzy" in fe_entry.flags and not extract_fuzzy:
                female = fe_entry.msgid
            else:
                female = fe_entry.msgstr

        resulting_entries.append({"index": file_index, "value": value, "female": female, "context": context})

    # combined occurrences may mess up order, restoring
    resulting_entries = sorted(resulting_entries, key=lambda k: k["index"])

    lines = []
    lines_female = []

    for res in resulting_entries:
        # get line format
        lfrm = get_line_format(res, ext)

        # add line to common/male package
        line = lfrm.format(index=res["index"], value=res["value"], context=res["context"], female=res["female"])
        # TODO: get rid of replace, handle improper characters in weblate
        lines.append(line.encode(encoding, "replace").decode(encoding))

        # add string to female package if needed
        if "female" in line_format and line_format["female"] == "separate":
            if res["female"] is not None:
                female_line = lfrm.format(index=res["index"], value=res["female"], context=res["context"])
            else:
                female_line = lfrm.format(index=res["index"], value=res["value"], context=res["context"])
            lines_female.append(female_line.encode(encoding, "replace").decode(encoding))

    # write main package
    with open(output_file, "w", encoding=encoding, newline=CONFIG.newline_tra) as file:
        file.writelines(lines)

    # explicitly disabled female?
    no_female = CONFIG.no_female

    if ("female" in line_format) and line_format["female"] == "separate" and dst_dir is not None and not no_female:
        # are translations the same? If yes, skipping copying "dialog" in sfall
        same = False
        if lines_female == lines:
            same = True

        # what's out path?
        female_file = get_female_filepath(output_file, dst_dir, same)

        # If need to create the file
        if same:  # if female translation is the same?
            if female_file is False:  # don't need to copy, automatic fallback
                print("  Female strings are same, not copying - sfall will fallback to male {}".format(output_file))
                return True  # cutoff the rest of the function
            else:
                print("  Female strings are same, copying to {}".format(female_file))
                copycreate(output_file, female_file)
        else:  # if it's different, extract separately
            if female_file is False:
                print(
                    "  WARN: female strings are different, but female file is not supported for path {}".format(
                        output_file
                    )
                )
                return True
            else:
                print("  Also extracting female counterpart into {}".format(female_file))
                create_dir(get_dir(female_file))  # create dir if not exists
                with open(female_file, "w", encoding=encoding, newline=CONFIG.newline_tra) as file2:
                    file2.writelines(lines_female)


# nasty hack for sfall's female strings placement
def get_female_filepath(path: str, dst_dir: str, same: bool = True):
    # default: just add _female suffix
    female_path = path.replace(dst_dir + os.sep, dst_dir + CONFIG.female_dir_suffix + os.sep)
    if CONFIG.extract_format == "sfall":
        female_path = False  # default for sfall: don't copy, it will fallback to male
        if "cuts" in path.split(os.sep):  # cuts dont' fallback
            female_path = path.replace(os.sep + "cuts" + os.sep, os.sep + "cuts_female" + os.sep)
        if "dialog" in path.split(os.sep) and not same:  # dialog, female translation differs
            female_path = path.replace(os.sep + "dialog" + os.sep, os.sep + "dialog_female" + os.sep)
    return female_path


def get_line_format(e, ext: str):
    """
    Takes translation entry in format {'index': index, 'value': value, 'female': female, 'context': context}
    And file extension
    Returns corresponding string with placeholders from line_format
    """
    ff = FILE_FORMAT[ext]
    line_format = ff["line_format"]
    if e["context"] is not None:  # entry with context
        lfrm = line_format["context"]
    elif (
        "female" in e and e["female"] is not None and "female" in line_format and line_format["female"] != "separate"
    ):  # format with native support for female strings
        lfrm = line_format["female"]
    else:  # no context and no female, or format without native support for female strings
        lfrm = line_format["default"]
    return lfrm


def copycreate(src_file, dst_file):
    dirname = os.path.dirname(dst_file)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    shutil.copyfile(src_file, dst_file)


def file2msgstr(
    input_file: str,
    po: polib.POFile,
    occurence_path: str,
    encoding=CONFIG.encoding,
    overwrite: bool = True,
    same: bool = False,
    female_map=None,
):
    """returns PO file object"""

    trans = TRANSFile(filepath=input_file, is_source=False, encoding=encoding)  # load translations

    # map entries to occurrences for faster access, part 1
    entries_dict = OrderedDict()
    for e in po:
        for eo in e.occurrences:
            entries_dict[(eo[0], eo[1])] = e
    if female_map is None:
        female_map = female_entries(po)

    for t in trans.entries:
        index = t.index
        value = t.value
        context = t.context
        female_value = t.female

        if (value is None) or (value == ""):
            print("WARN: no msgid found for {}:{}, skipping string\n      {}".format(occurence_path, index, value))
            continue

        new_female_enties = []
        if (occurence_path, index) in entries_dict:
            # map entries to occurrences for faster access, part 2
            e: polib.POEntry = entries_dict[(occurence_path, index)]

            if female_value:
                # female entries have no occurences, checking if female translation already present
                if e.msgid in female_map:
                    fe: polib.POEntry = female_map[e.msgid]
                    if fe and (fe.msgstr != female_value):
                        print("INFO: female translation change detected:")
                        print("  ORIG: {}".format(e.msgid))
                        print("  OLD:  {}".format(fe.msgstr))
                        print("  NEW:  {}".format(female_value))
                        skip = False
                        if not overwrite:
                            print("  Female translation already exists, overwrite disabled, skipping")
                            skip = True
                        if not skip and (e.msgid == female_value):
                            if same:
                                print("INFO: source and female translation are the same. Using it regardless.")
                                print("   {}".format(e.msgid))
                                print("   {}".format(female_value))
                            else:
                                print(
                                    "INFO: source and female translation are the same for {}. Skipping:".format(
                                        e.occurrences
                                    )
                                )
                                print("   {}".format(e.msgid))
                                print("   {}".format(female_value))
                                skip = True
                        if not skip:
                            fe.msgstr = female_value
                            if "fuzzy" in fe.flags:
                                print("    Unfuzzied female entry")
                                fe.flags.remove("fuzzy")
                                fe.previous_msgid = None
                elif e.msgstr != female_value:
                    print("INFO: new female translation detected:")
                    print("  ORIG:   {}".format(e.msgid))
                    print("  MALE:   {}".format(e.msgstr))
                    print("  FEMALE: {}".format(female_value))
                    fe = polib.POEntry(msgid=e.msgid, msgstr=female_value, msgctxt=CONTEXT_FEMALE)
                    new_female_enties.append(fe)

            # translation is the same
            if e.msgstr == value and e.msgctxt == context:
                print("  translation is the same for {}".format(e.occurrences))
                if "fuzzy" in e.flags:
                    if CONFIG.extract_fuzzy:
                        print("  {}  is fuzzy. Keeping fuzzy flag.".format(e.occurrences))
                        continue
                    else:
                        print(
                            f"  {e.occurrences} is fuzzy, but extract_fuzzy is not set. "
                            "Assuming manual translation change to the same value, clearing fuzzy flag."
                        )

            # translation is the same as source
            if e.msgid == value and not same:
                print("INFO: string and new translation are the same for {}. Skipping:".format(e.occurrences))
                print("   {}".format(e.msgid))
                continue

            # if translation already exists and different
            if e.msgstr is not None and e.msgstr != "" and e.msgstr != value:
                # if overwrite is disabled, cutoff
                if not overwrite:
                    print(
                        "INFO: translation already exists for {}, overwrite disabled, skipping:".format(e.occurrences)
                    )
                    continue

            # finally, all checks passed
            print("INFO: translation update found for {}.".format(e.occurrences))
            print("  Replacing old string with new:")
            print("    ORIG: {}".format(e.msgid))
            print("    OLD:  {}".format(e.msgstr))
            print("    NEW:  {}".format(value))
            e.msgstr = value
            e.msgctxt = context
            if "fuzzy" in e.flags or e.previous_msgid:
                print("    Unfuzzied entry")
                e.previous_msgid = None
                if "fuzzy" in e.flags:
                    e.flags.remove("fuzzy")

    # add newly found female entries
    for nfe in new_female_enties:
        po.append(nfe)

    return po


# check if TXT file is indexed
def is_indexed(txt_filename: str, encoding=CONFIG.encoding):
    f = open(txt_filename, "r", encoding=encoding)
    # count non-empty lines
    num_lines = sum(1 for line in f if line.rstrip())
    f.close()

    # count lines that are indexed
    pattern = FILE_FORMAT["txt"]["pattern"]
    f = open(txt_filename, "r", encoding=encoding)
    text = f.read()
    indexed_lines = re.findall(pattern, text)
    num_indexed_lines = len(indexed_lines)
    f.close()
    if num_lines == num_indexed_lines:
        return True
    else:
        return False


def sort_po(po: polib.POFile):
    for e in po:
        e.occurrences = natsorted(e.occurrences, key=lambda k: (k[0], k[1]))
    metadata = po.metadata
    po = natsorted(
        po, key=lambda k: k.occurrences[0] if len(k.occurrences) > 0 else ("zzzzz", "99999")
    )  # female empty occurences hack
    po2 = polib.POFile()
    po2.metadata = metadata
    po2.extend(po)
    return po2


def po_make_unique(po):
    entries_dict = OrderedDict()
    old_metadata = po.metadata
    for e in po:
        if (e.msgid, e.msgctxt) in entries_dict:

            e0 = entries_dict[(e.msgid, e.msgctxt)]
            e0.occurrences.extend(e.occurrences)

            if e.comment is not None:
                if e0.comment is None:
                    e0.comment = e.comment
                elif e0.comment != e.comment:
                    e0.comment = e0.comment + "; " + e.comment

            if e.tcomment is not None:
                if e0.tcomment is None:
                    e0.tcomment = e.tcomment
                elif e0.tcomment != e.tcomment:
                    e0.tcomment = e0.tcomment + "; " + e.tcomment

            for f in e.flags:
                if f not in e0.flags:
                    e0.flags.append(f)

            if e.previous_msgctxt and not e0.previous_msgctxt:
                e0.previous_msgctxt = e.previous_msgctxt
            if e.previous_msgid and not e0.previous_msgid:
                e0.previous_msgid = e.previous_msgid
            if e.previous_msgid_plural and not e0.previous_msgid_plural:
                e0.previous_msgid_plural = e.previous_msgid_plural

        else:
            entries_dict[(e.msgid, e.msgctxt)] = e
    po2 = polib.POFile()
    po2.metadata = old_metadata
    for key, value in list(entries_dict.items()):
        po2.append(value)
    return po2


class TRANSEntry:
    def __init__(self):
        self.index = None
        self.value = None
        self.context = None
        self.female = None
        self.comment = None
        self.occurence = None


class TRANSFile:
    """
    Common translation class, holding translation entries of a single file
    is_source: if set, adds EMPTY_COMMENT to all empty lines
    This is because PO gettext format doesn't tolerate empty msgids
    """

    def __init__(self, filepath: str, is_source: False, encoding=CONFIG.encoding):
        self.entries: list[TRANSEntry] = []
        self.encoding = encoding
        fext = get_ext(filepath)
        self.fformat = FILE_FORMAT[fext]
        self.pattern = self.fformat["pattern"]
        self.dotall = self.fformat["dotall"]
        self.filepath = filepath

        try:  # comment for all entries in file
            self.comment = self.fformat["comment"]
        except:
            pass

        self.lines = self.load_lines(filepath)

        # enabled for file2msgstr, disabled for file2po
        if not is_source:
            self.lines_female = None
            if self.fformat["line_format"]["female"] == "separate":
                female_dir = get_dir(filepath) + CONFIG.female_dir_suffix
                female_file = os.path.join(female_dir, basename(filepath))
                print("  separate file format, looking for female file {}... ".format(female_file), end="")
                if os.path.isfile(female_file):
                    print("found")
                    self.lines_female = self.load_lines(female_file)
                else:
                    print("didn't find")

            if self.lines_female is not None:
                if self.lines_female == self.lines:
                    print("  female lines are identical")
                else:
                    print("  female lines are different")

        # protection again duplicate indexes, part 1
        seen = []

        for line in self.lines:
            entry = TRANSEntry()

            # index and value
            index = line[self.fformat["index"]]
            entry.occurence = (filepath, str(index))
            entry.value = str(line[self.fformat["value"]])

            # skip invalid '000' entries in MSG files
            if fext == "msg" and index == "000":
                print(
                    "WARN: {} - invalid entry number found, skipping:\n     {{000}}{{}}{{{}}}".format(
                        filepath, entry.value
                    )
                )
                continue

            entry.index = line[self.fformat["index"]]

            # comment
            # 1. generic comment for all entries in file
            try:
                entry.comment = self.fformat["comment"]
            except:
                pass
            # 2. handle empty lines in source files
            if entry.value == "":
                if is_source is True:
                    entry.value = " "
                    entry.comment = EMPTY_COMMENT

            # context
            try:
                entry.context = line[self.fformat["context"]]
            except:
                pass
            if entry.context == "":
                entry.context = None

            # female
            if fext == "tra":  # TRA file specific
                try:
                    entry.female = str(line[self.fformat["female"]])
                except:
                    pass
                if entry.female == "":
                    entry.female = None

                if entry.female and entry.context:
                    print("ERROR. TRA strings with female variants may not have context.")
                    print(line)
                    print(entry)
                    sys.exit(1)

            # sfall female extraction
            if not is_source and (self.lines_female is not None) and not (self.lines_female == self.lines):
                try:
                    female_line = [fl for fl in self.lines_female if fl[self.fformat["index"]] == entry.index][0]
                    entry.female = str(female_line[self.fformat["value"]])
                    if entry.female != entry.value:
                        print("  found alternative female string for line {}: {}".format(entry.index, entry.female))
                except:
                    pass

            # protection against duplicate indexes, part 2
            if entry.index in seen:
                print(
                    "WARN: duplicate string definition found {}:{}, using new value:\n      {}".format(
                        filepath, entry.index, entry.value
                    )
                )
                self.entries = [entry if x.index == entry.index else x for x in self.entries]
                continue
            else:
                seen.append(index)

            # produce the final list of strings
            if entry.value is not None and entry.value != "":
                self.entries.append(entry)

    def load_lines(self, filepath: str):
        with open(filepath, "r", encoding=self.encoding) as fh:
            text = fh.read()
            if self.dotall:
                lines = re.findall(self.pattern, text, re.DOTALL)
            else:
                lines = re.findall(self.pattern, text)
        return lines


def language_slug(po_filename):
    """
    Allows to extract PO files into simplified language names: pt_BR.po -> portuguese/1.msg.
    Working with language codes is not convenient in mods.
    A temporary hack until a better solution is found.
    """
    slug_map = {
        "de": "german",
        "fr": "french",
        "pt_br": "portuguese",
        "it": "italian",
        "pl": "polish",
        "es": "spanish",
        "ru": "russian",
        "sv": "swedish",
        "uk": "ukrainian",
    }
    slug = strip_ext(basename(po_filename)).lower()
    if CONFIG.simple_languages:
        try:
            slug = slug_map[slug]
        except:
            pass
    return slug


def update_female_entries(po: polib.POFile):
    """
    (Un)obsoletes and if necessary (un)fuzzies female strings that have a corresponding male counterpart.
    (Male = no context)
    """
    # this also includes fuzzies
    male_entries = {x.msgid: x for x in po if ((x.msgctxt != CONTEXT_FEMALE) and (not x.obsolete))}
    # and this is for matching male strings that were changed
    fuzzy_male_entries = {
        x.previous_msgid: x
        for x in po
        if ((x.previous_msgid is not None) and (x.msgctxt != CONTEXT_FEMALE) and (not x.obsolete))
    }

    for e in po.obsolete_entries():
        if e.msgctxt != CONTEXT_FEMALE:
            continue

        # if exact male match found, unobsolete
        if e.msgid in male_entries:
            male_entry = male_entries[e.msgid]
            # if it's fuzzy, doing the same for female
            e.previous_msgid = male_entry.previous_msgid
            e.flags = male_entry.flags
            e.obsolete = False

        # else, check if a fuzzy male match exists, and fix female to have the same attributes
        elif e.msgid in fuzzy_male_entries:
            male_entry = fuzzy_male_entries[e.msgid]
            e.msgid = male_entry.msgid
            e.previous_msgid = male_entry.previous_msgid
            if "fuzzy" not in e.flags:
                e.flags.append("fuzzy")
            e.obsolete = False

    # and delete female entries with no non-obsolete male match
    new_entries = [x for x in po if not (x.obsolete and (x.msgctxt == CONTEXT_FEMALE))]
    meta = po.metadata
    po2 = polib.POFile()
    po2.metadata = meta
    for e in new_entries:
        po2.append(e)
    return po2


def unfuzzy_exact_matches(po: polib.POFile):
    """
    For some reason msgmerge won't clear fuzzy flag if source string is changed, then changed back:
    #: game/g_map_hotkey.msg:1
    #, fuzzy
    #| msgid "Nah... I think I'm gonna take the ladder."
    msgid "Nah... I think I'm gonna take the ladder."
    msgstr "Naa... Penso che prendero' la scala."

    This function unfuzzies such entries.
    """
    for e in po.fuzzy_entries():
        if (e.previous_msgid == e.msgid) and (e.previous_msgctxt == e.msgctxt):
            print("    Unfuzzied entry {}, exact match with previous".format(e.occurrences))
            e.flags.remove("fuzzy")
            e.previous_msgid = None
            e.previous_msgctxt = None
    return po


class LanguageMap:
    """
    map PO basenames to language dir names, for unpoify and dir2msgstr
    This is because languages added through Weblate use codes like pt_BR, see find_files
    """

    def __init__(self):
        po_files = find_files(CONFIG.po_dir, "po")
        self.slug2po = {}
        self.po2slug = {}
        for pf in po_files:
            pf = basename(pf)
            slug = language_slug(pf)
            self.slug2po[slug] = pf
            self.po2slug[pf] = slug
